﻿using System;
using System.ComponentModel;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Media;
using System.Windows.Navigation;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Microsoft.Phone.Tasks;
using Petzold.Phone.EBookReader;

namespace HorrorReader
{
    public partial class BookViewerPage : PhoneApplicationPage
    {
        AppSettings appSettings = (Application.Current as App).AppSettings;
        WebBrowserTask webBrowserTask = new WebBrowserTask();
        SearchTask searchTask = new SearchTask();
        BookInfo book;
        PageProvider pageProvider;
        Popup activePopup;

        // Three fields consolidated for tombstoning purposes
        public struct FindParams
        {
            public bool FirstFindFromTop { set; get; }
            public int ParagraphIndex { set; get; }
            public int CharacterIndex { set; get; }
        }
        FindParams findParams;

        // Also for tombstoning purposes
        FindInfo currentHighlight;

        public BookViewerPage()
        {
            InitializeComponent();

            // Set the fields of the appbar buttons and menu
            appbar = this.ApplicationBar as ApplicationBar;

            chaptersIconButton = appbar.Buttons[0] as ApplicationBarIconButton;
            bookmarksIconButton = appbar.Buttons[1] as ApplicationBarIconButton;
            annotationsIconButton = appbar.Buttons[2] as ApplicationBarIconButton;
            findIconButton = appbar.Buttons[3] as ApplicationBarIconButton;

            smallerTextSize = this.ApplicationBar.MenuItems[1] as ApplicationBarMenuItem;
            largerTextSize = this.ApplicationBar.MenuItems[2] as ApplicationBarMenuItem;

            Loaded += OnLoaded;
        }

        protected override void OnNavigatedFrom(NavigationEventArgs args)
        {
            if (activePopup != null)
            {
                this.State["popup"] = activePopup.Name;
                this.State["popupOffsetX"] = activePopup.HorizontalOffset;
                this.State["popupOffsetY"] = activePopup.VerticalOffset;
                this.State["findParams"] = findParams;

                if (currentHighlight != null)
                {
                    this.State["currentHighlight"] = currentHighlight;
                }
                else if (this.State.ContainsKey("currentHighlight"))
                {
                    this.State.Remove("currentHighlight");
                }

                ITombstonable popupChild = activePopup.Child as ITombstonable;

                if (popupChild != null)
                    popupChild.SaveState(activePopup.Name, this.State);
            }
            else if (this.State.ContainsKey("popup"))
            {
                this.State.Remove("popup");
            }

            base.OnNavigatedFrom(args);
        }

        protected override void OnNavigatedTo(NavigationEventArgs args)
        {
            // Only reload the book and recreate the PageProvider if there's really
            //  been some tombstoning, and not when navigating to another page
            //  in the app.
            if (pageProvider == null)
            {
                book = (Application.Current as App).CurrentBook;
                pageProvider = new PageProvider(book);
                bookViewer.PageProvider = pageProvider;
                this.DataContext = book;
            }
            if (activePopup == null && this.State.ContainsKey("popup"))
            {
                EnableViewerAndAppBar(false);
                activePopup = this.FindName(this.State["popup"] as string) as Popup;
                activePopup.HorizontalOffset = (double)this.State["popupOffsetX"];
                activePopup.VerticalOffset = (double)this.State["popupOffsetY"];
                findParams = (FindParams)this.State["findParams"];
                ITombstonable popupChild = activePopup.Child as ITombstonable;

                if (this.State.ContainsKey("currentHighlight"))
                {
                    currentHighlight = (FindInfo)this.State["currentHighlight"];
                }

                if (popupChild != null)
                    popupChild.RestoreState(activePopup.Name, this.State);

                activePopup.IsOpen = true;
            }
            else
            {
                EnableViewerAndAppBar(true);
            }

            base.OnNavigatedTo(args);
        }

        // Occurs after OnNavigatedTo
        void OnLoaded(object sender, RoutedEventArgs args)
        {
            switch (appSettings.PageTransition)
            {
                case "slide": bookViewer.PageTransition = new SlideTransition(); break;
                case "skew": bookViewer.PageTransition = new SkewTransition(); break;
                case "curl": bookViewer.PageTransition = new CurlTransition(); break;
                case "flip": bookViewer.PageTransition = new FlipTransition(); break;
            }
            
            ResetBookViewerInformationAndGotoPage();

            if (currentHighlight != null)
                bookViewer.HighlightText(currentHighlight);
        }

        protected override void OnBackKeyPress(CancelEventArgs args)
        {
            if (activePopup != null)
            {
                activePopup.IsOpen = false;
                activePopup = null;
                EnableViewerAndAppBar(true);
                args.Cancel = true;
            }

            base.OnBackKeyPress(args);
        }

        void OnBookViewerPageChanged(object sender, EventArgs args)
        {
            int chapterIndex = bookViewer.ChapterIndex;
            int pageIndex = bookViewer.PageIndex;

            if (chapterIndex < book.Chapters.Count && pageIndex < book.Chapters[chapterIndex].Pages.Count)
            {
                int accumulatedCharacterCount = book.Chapters[chapterIndex].Pages[pageIndex].AccumulatedCharacterCount;

                // Save for isolated storage settings
                book.CurrentChapterIndex = chapterIndex;
                book.CurrentPageIndex = pageIndex;
                book.CurrentCharacterOffset = accumulatedCharacterCount;

                // Update the percentage indicator
                double percentage = 100.0 * accumulatedCharacterCount / book.TotalCharacterCount;
                percentageText.Text = String.Format("{0:F1}%", percentage);
            }
        }

        void ResetBookViewerInformationAndGotoPage()
        {
            // Set the BookViewer font information
            bookViewer.FontFamily = new FontFamily(appSettings.FontFamily);
            bookViewer.FontSize = (double)this.Resources["PhoneFontSizeNormal"] * appSettings.FontSizeMultiplier;

            // Set the pageProvider font information; could cause purging of pagination info
            pageProvider.SetFont(bookViewer.FontFamily, bookViewer.FontSize);

            // This is also a good place to enable the items on the menu
            largerTextSize.IsEnabled = appSettings.FontSizeMultiplier < AppSettings.MaximumFontSizeMultiplier;
            smallerTextSize.IsEnabled = appSettings.FontSizeMultiplier > AppSettings.MinimumFontSizeMultiplier;

            // If page is expanded, animate it down
            bookViewer.AnimatePinchAndPanUndo();

            if (book.IsChapterAndPageInvalid)
            {
                bookViewer.JumpToCharacterOffset(book.CurrentCharacterOffset);
                book.IsChapterAndPageInvalid = false;                               // set here or elsewhere????
            }

            // Depending on retention of pagination information, one or the other of these will go to the page
            else if (book.Chapters[book.CurrentChapterIndex].Pages.Count > book.CurrentPageIndex)
            {
                if (!bookViewer.JumpToPage(book.CurrentChapterIndex, book.CurrentPageIndex))
                    bookViewer.JumpToCharacterOffset(book.CurrentCharacterOffset);
            }
            else
            {
                bookViewer.JumpToCharacterOffset(book.CurrentCharacterOffset);
            }
        }

        void OnAppBarChaptersButtonClick(object sender, EventArgs args)
        {
            // Navigate to the Chapters page
            this.NavigationService.Navigate(new Uri("/ChaptersPage.xaml", UriKind.Relative));
        }

        void OnAppBarBookmarksButtonClick(object sender, EventArgs args)
        {
            // Navigate to the Bookmarks page
            this.NavigationService.Navigate(new Uri("/BookmarksPage.xaml", UriKind.Relative));
        }

        void OnAppBarAnnotationsButtonClick(object sender, EventArgs args)
        {
            // Navigate to the Annotations page
            this.NavigationService.Navigate(new Uri("/AnnotationsPage.xaml", UriKind.Relative));
        }

        void OnAppBarSearchButtonClick(object sender, EventArgs args)
        {
            bookViewer.AnimatePinchAndPanUndo();

            ChapterInfo chapter = book.Chapters[book.CurrentChapterIndex];
            PageInfo page = chapter.Pages[book.CurrentPageIndex];
            findParams.ParagraphIndex = page.ParagraphIndex;
            findParams.CharacterIndex = page.CharacterIndex;
            findParams.FirstFindFromTop = true;

            EnableViewerAndAppBar(false);
            findDialogPopup.IsOpen = true;
            findDialogPopup.VerticalOffset = 0;
            activePopup = findDialogPopup;
        }

        void OnAppBarSettingsItemClick(object sender, EventArgs args)
        {
            // Navigate to Settings page
            string uri = String.Format("/SettingsPage.xaml");
            this.NavigationService.Navigate(new Uri(uri, UriKind.Relative));
        }

        void OnAppBarTextSizeItemClick(object sender, EventArgs args)
        {
            if (sender == smallerTextSize)
            {
                appSettings.FontSizeMultiplier = Math.Max(AppSettings.MinimumFontSizeMultiplier,
                                                          appSettings.FontSizeMultiplier - 0.1);
            }
            else if (sender == largerTextSize)
            {
                appSettings.FontSizeMultiplier = Math.Min(AppSettings.MaximumFontSizeMultiplier,
                                                          appSettings.FontSizeMultiplier + 0.1);
            }

            ResetBookViewerInformationAndGotoPage();
        }

        // When a TextSelected event comes from the BookViewer control,
        //  disable the control and display the text-selection menu.
        void OnBookViewerTextSelected(object sender, TextSelectedEventArgs args)
        {
            // Save highlight for tombstoning purposes
            currentHighlight = new FindInfo
            {
                ParagraphIndex = args.ParagraphIndex,
                StartCharacterIndex = args.CharacterIndex,
                EndCharacterIndex = args.CharacterIndex + args.CharacterCount
            };

            EnableViewerAndAppBar(false);

            textSelectionMenu.Tag = args;
            textSelectionMenu.SelectedItem = null;
            textSelectionMenuPopup.IsOpen = true;
            activePopup = textSelectionMenuPopup;

            // Position the menu
            textSelectionMenu.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
            Size selectionMenuSize = textSelectionMenu.DesiredSize;
            double offsetX = 0;
            double offsetY = 0;

            // Try centering it underneath the text
            if (args.Bounds.Bottom + selectionMenuSize.Height < ContentPanel.ActualHeight)
            {
                offsetX = (args.Bounds.Left + args.Bounds.Right - selectionMenuSize.Width) / 2;
                offsetY = args.Bounds.Bottom;
            }
            // Now try centering it above the text
            else if (args.Bounds.Top - selectionMenuSize.Height > 0)
            {
                offsetX = (args.Bounds.Left + args.Bounds.Right - selectionMenuSize.Width) / 2;
                offsetY = args.Bounds.Top - selectionMenuSize.Height;
            }
            // How about at the right?
            else if (args.Bounds.Right + selectionMenuSize.Width < ContentPanel.ActualWidth)
            {
                offsetX = args.Bounds.Right;
                offsetY = (args.Bounds.Top + args.Bounds.Bottom - selectionMenuSize.Height) / 2;
            }
            // Maybe at the left will be OK
            else if (args.Bounds.Left - selectionMenuSize.Width > 0)
            {
                offsetX = args.Bounds.Left - selectionMenuSize.Width;
                offsetY = (args.Bounds.Top + args.Bounds.Bottom - selectionMenuSize.Height) / 2;
            }
            // I guess we have to stick it in a corner
            else
            {
                if (args.Bounds.Left > ContentPanel.ActualWidth - args.Bounds.Right)
                    offsetX = 0;
                else
                    offsetX = ContentPanel.ActualWidth - selectionMenuSize.Width;

                if (args.Bounds.Top > ContentPanel.ActualHeight - args.Bounds.Bottom)
                    offsetY = 0;
                else
                    offsetY = ContentPanel.ActualHeight - selectionMenuSize.Height;
            }

            // Finally, make sure it's on the screen
            offsetX = Math.Max(0, Math.Min(offsetX, ContentPanel.ActualWidth - selectionMenuSize.Width));
            offsetY = Math.Max(0, Math.Min(offsetY, ContentPanel.ActualHeight - selectionMenuSize.Height));

            textSelectionMenuPopup.HorizontalOffset = offsetX;
            textSelectionMenuPopup.VerticalOffset = offsetY;
        }

        void OnTextSelectionMenuSelectionChanged(object sender, SelectionChangedEventArgs args)
        {
            ListBox lstbox = sender as ListBox;
            TextSelectedEventArgs selectionArgs = lstbox.Tag as TextSelectedEventArgs;

            if (lstbox.SelectedItem != null)
            {
                textSelectionMenuPopup.IsOpen = false;
                activePopup = null;
                EnableViewerAndAppBar(true);
                string command = (lstbox.SelectedItem as ListBoxItem).Content as string;
                string text = selectionArgs.Text;

                switch (command)
                {
                    case "note":
                        int characterOffset = book.Paragraphs[selectionArgs.ParagraphIndex].AccumulatedCharacterCount +
                                                    selectionArgs.CharacterIndex;

                        Annotation annotation = new Annotation
                        {
                            Text = selectionArgs.Text,
                            Note = "",

                            ChapterTitle = book.Chapters[book.CurrentChapterIndex].Title,
                            CharacterOffset = characterOffset,
                            Percentage = 100.0 * characterOffset / book.TotalCharacterCount,
                            DateTime = DateTime.Now,
                        };
                        annotationDialog.Annotation = annotation;
                        annotationDialogPopup.IsOpen = true;
                        activePopup = annotationDialogPopup;
                        EnableViewerAndAppBar(false);
                        break;

                    case "bing":
                        searchTask.SearchQuery = StripPunctuation(text);
                        searchTask.Show();
                        break;

                    case "dictionary":
                        InvokeWebBrowser("http://www.google.com/dictionary?langpair=en|en&q={0}", text);
                        break;

                    case "wikipedia":
                        InvokeWebBrowser("http://www.wikipedia.org/wiki/{0}", text);
                        break;

                    case "cancel":
                        break;
                }
            }
        }

        void InvokeWebBrowser(string format, string text)
        {
            string url = String.Format(format, StripPunctuation(text));
            webBrowserTask.URL = Uri.EscapeDataString(url);
            webBrowserTask.Show();
        }

        string StripPunctuation(string text)
        {
            StringBuilder strBuilder = new StringBuilder();

            foreach (char ch in text)
                if (!Char.IsPunctuation(ch) || ch == '-')
                    strBuilder.Append(ch);

            return strBuilder.ToString();
        }

        void OnAnnotationDialogDismissed(object sender, CancelEventArgs args)
        {
            AnnotationDialog annotationDialog = sender as AnnotationDialog;

            if (!args.Cancel)
            {
                book.Annotations.Add(annotationDialog.Annotation);
            }

            annotationDialogPopup.IsOpen = false;
            activePopup = null;
            EnableViewerAndAppBar(true);
        }

        void OnBookViewerBookmarkInserted(object sender, EventArgs args)
        {
            Bookmark bookmark = new Bookmark
            {
                Label = "bookmark" + (book.Bookmarks.Count + 1),
                ChapterTitle = book.Chapters[book.CurrentChapterIndex].Title,
                CharacterOffset = book.CurrentCharacterOffset,
                Percentage = 100.0 * book.CurrentCharacterOffset / book.TotalCharacterCount,
                DateTime = DateTime.Now,
            };

            bookmarkDialog.Bookmark = bookmark;
            EnableViewerAndAppBar(false);
            bookmarkDialogPopup.IsOpen = true;
            activePopup = bookmarkDialogPopup;
        }

        void OnBookmarkDialogDismissed(object sender, CancelEventArgs args)
        {
            BookmarkDialog bookmarkDialog = sender as BookmarkDialog;

            if (!args.Cancel)
            {
                book.Bookmarks.Add(bookmarkDialog.Bookmark);
            }

            bookmarkDialogPopup.IsOpen = false;
            activePopup = null;
            EnableViewerAndAppBar(true);
        }

        void OnFindDialogFindText(object sender, FindTextEventArgs args)
        {
            // Avoid refinding the previous find
            if (!findParams.FirstFindFromTop && !args.GoUp)
                findParams.CharacterIndex += 1;

            FindInfo findInfo = pageProvider.FindText(findParams.ParagraphIndex, findParams.CharacterIndex, 
                                                      args.Text, args.MatchCase, args.WholeWord, args.GoUp);
            bookViewer.ClearHighlighting();
            currentHighlight = null;

            if (findInfo == null)
            {
                FindDialog findDialog = sender as FindDialog;
                findDialog.ShowNoMoreFinds();
            }
            else
            {
                bookViewer.JumpToCharacterOffset(findInfo.CharacterOffset);
                bookViewer.HighlightText(findInfo);
                currentHighlight = findInfo;

                if (findInfo.WordLocation < findDialog.ActualHeight)
                    findDialogPopup.VerticalOffset = bookViewer.ActualHeight - findDialog.ActualHeight
                                                        - findDialog.Margin.Bottom - findDialog.Margin.Top;
                else
                    findDialogPopup.VerticalOffset = 0;

                findParams.ParagraphIndex = findInfo.ParagraphIndex;
                findParams.CharacterIndex = findInfo.StartCharacterIndex;
            }

            findParams.FirstFindFromTop = false;
        }

        void OnFindDialogDialogDismissed(object sender, CancelEventArgs args)
        {
            findDialogPopup.IsOpen = false;
            activePopup = null;
            EnableViewerAndAppBar(true);
        }

        void EnableViewerAndAppBar(bool isEnabled)
        {
            // On transition from disabled BookViewer to enabled BookViewer, clear highlighting
            if (isEnabled && !bookViewer.IsHitTestVisible)
            {
                bookViewer.ClearHighlighting();
                currentHighlight = null;
            }

            bookViewer.IsHitTestVisible = isEnabled;
            chaptersIconButton.IsEnabled = isEnabled;
            bookmarksIconButton.IsEnabled = isEnabled && book.Bookmarks.Count > 0;
            annotationsIconButton.IsEnabled = isEnabled && book.Annotations.Count > 0;
            findIconButton.IsEnabled = isEnabled;
            appbar.IsMenuEnabled = isEnabled;
        }
    }
}