﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using Microsoft.Phone.Controls;

namespace Petzold.Phone.EBookReader
{
    public partial class BookViewer
    {
        // Gesture control
        bool gestureAllowed, canGoNextPage, canGoPrevPage, gestureHadFlick;
        ViewerDisplayMode viewerDisplayMode = ViewerDisplayMode.Normal;
        ViewerTouchMode viewerTouchMode = ViewerTouchMode.Reading;

        // Field for dragging pages
        double dragFactor;

        // Field for pinch-and-pan
        Point ptPinchStart;

        // Fields for text selection
        int selectedParagraphIndex;
        TextBlock selectedTextBlock1, selectedTextBlock2;

        void OnGestureListenerGestureBegin(object sender, GestureEventArgs args)
        {
            // Initialize all four Boolean fields
            gestureAllowed = viewerTouchMode == ViewerTouchMode.Reading;    // ie, != ViewerTouchMode.Animating

            if (!gestureAllowed)
                return;

            gestureHadFlick = false;
            canGoNextPage = pageHosts[(pageHostBaseIndex + 1) % 3].Child != null;
            canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;
            args.Handled = true;
        }

        void OnGestureListenerTap(object sender, GestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            // If we're in a zoomed state, undo the zoom
            if (viewerDisplayMode == ViewerDisplayMode.Zoomed)
            {
                AnimatePinchAndPanUndo();
            }
            // Otherwise go to the next page if possible
            else if (canGoNextPage)
            {
                AnimateToNextPage();
            }
            args.Handled = true;
        }

        void OnGestureListenerHold(object sender, GestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            // For Hold, switch to Selecting mode
            if (viewerTouchMode == ViewerTouchMode.Reading &&     // Won't be anything else, actually
                args.OriginalSource is TextBlock)
            {
                // Get the TextBlock underneath the finger
                TextBlock txtblk = args.OriginalSource as TextBlock;
                WordInfo word = txtblk.Tag as WordInfo;

                // This means it's part of the text
                if (word != null)
                {
                    // Initiate selection mode
                    selectedParagraphIndex = word.ParagraphIndex;
                    selectedTextBlock1 = txtblk;
                    selectedTextBlock2 = txtblk;
                    SelectTextBlockRange(selectedTextBlock1, selectedTextBlock2, true);

                    viewerTouchMode = ViewerTouchMode.Selecting;
                }
            }
            args.Handled = true;
        }

        void OnGestureListenerFlick(object sender, FlickGestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            // Zip over to the edge
            if (viewerDisplayMode == ViewerDisplayMode.Zoomed)
            {
                ConsolidateTransforms();
                double translateX = 0;
                double translateY = 0;

                if (args.Angle < 45)        // flick to the right
                    translateX += 100000;
                else if (args.Angle < 135)  // flick to the bottom
                    translateY += 100000;
                else if (args.Angle < 225)  // flick to the left
                    translateX -= 100000;
                else if (args.Angle < 315)  // flick to the top
                    translateY -= 100000;
                else                        // flick to the right
                    translateX += 100000;

                ClampPinchAndPanTransforms(1, translateX, translateY);
                ConsolidateTransforms();
            }
            // This is the logic for modes of Reading (a Flick w/o a Drag)
            //      or Dragging (a Flick after a Drag)
            else if (viewerTouchMode != ViewerTouchMode.Selecting)
            {
                if (args.Angle < 45 || args.Angle > 315)
                {
                    if (!canGoPrevPage)
                    {
                        if (this.ChapterIndex != 0 && this.PageIndex == 0)
                        {
                            pageHosts[(pageHostBaseIndex + 2) % 3].Child = PageProvider.GetLastPage(prevPageChapterIndex, out prevPagePageIndex);
                            canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;
                        }
                    }
                    if (canGoPrevPage)
                    {
                        gestureHadFlick = true;
                        AnimateToPreviousPage();
                    }
                }

                else if (args.Angle > 135 && args.Angle < 225 && canGoNextPage)
                {
                    gestureHadFlick = true;
                    AnimateToNextPage();
                }
                // Generate BookmarkInserted event
                else
                {
                    if (BookmarkInserted != null)
                        BookmarkInserted(this, EventArgs.Empty);
                }
            }
            args.Handled = true;        
        }

        void OnGestureListenerDragStarted(object sender, DragStartedGestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            if (viewerTouchMode == ViewerTouchMode.Selecting)
            {
            }
            else if (viewerDisplayMode == ViewerDisplayMode.Zoomed)
            {
                viewerTouchMode = ViewerTouchMode.Panning;
            }
            else if (viewerTouchMode == ViewerTouchMode.Reading)
            {
                viewerTouchMode = ViewerTouchMode.Dragging;
                dragFactor = 0;
            }
            args.Handled = true;
        }

        void OnGestureListenerDragDelta(object sender, DragDeltaGestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            switch (viewerTouchMode)
            {
                case ViewerTouchMode.Dragging:
                    double newDragFactor = dragFactor - args.HorizontalChange / pageHosts[pageHostBaseIndex].ActualWidth;

                    if (newDragFactor < 0 && !canGoPrevPage && this.ChapterIndex != 0 && this.PageIndex == 0)
                    {
                        pageHosts[(pageHostBaseIndex + 2) % 3].Child = PageProvider.GetLastPage(prevPageChapterIndex, out prevPagePageIndex);
                        canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;
                    }

                    double maximum = canGoNextPage ? 1 : 0;
                    double minimum = canGoPrevPage ? -1 : 0;
                    dragFactor = Math.Max(minimum, Math.Min(maximum, newDragFactor));
                    this.PageTransition.FractionalBaseIndex = pageHostBaseIndex + dragFactor;
                    break;

                case ViewerTouchMode.Selecting:
                    // Find TextBlock under touch point
                    Point pt = args.GetPosition(Application.Current.RootVisual);
                    IEnumerable<UIElement> hitTestElements = VisualTreeHelper.FindElementsInHostCoordinates(pt, pageHosts[pageHostBaseIndex]);

                    foreach (UIElement hitTestElement in hitTestElements)
                    {
                        if (hitTestElement is TextBlock)
                        {
                            TextBlock txtblk = hitTestElement as TextBlock;
                            WordInfo word = txtblk.Tag as WordInfo;

                            // Extend the selection to that word
                            if (word != null && word.ParagraphIndex == selectedParagraphIndex)
                            {
                                selectedTextBlock2 = txtblk;
                                SelectTextBlockRange(selectedTextBlock1, selectedTextBlock2, true);
                                break;
                            }
                        }
                    }
                    break;

                case ViewerTouchMode.Panning:
                    // Set translation transform only
                    double translateX = translateTransform.X + args.HorizontalChange;
                    double translateY = translateTransform.Y + args.VerticalChange;
                    ClampPinchAndPanTransforms(1, translateX, translateY);
                    break;
            }
            args.Handled = true;
        }

        void OnGestureListenerDragCompleted(object sender, DragCompletedGestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            switch (viewerTouchMode)
            {
                // This logic is in GestureCompleted
                case ViewerTouchMode.Dragging:
                case ViewerTouchMode.Selecting:
                    break;

                case ViewerTouchMode.Panning:
                    // Transfer Scale and Translate transforms to Matrix transform
                    ConsolidateTransforms();
                    break;
            }
            args.Handled = true;
        }

        void OnGestureListenerPinchStarted(object sender, PinchStartedGestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            if (viewerTouchMode == ViewerTouchMode.Reading)
            {
                // Find scaling center, subject to current transform
                Point ptScaleCenter = args.GetPosition(manipulationGrid);
                ptScaleCenter = matrixTransform.Transform(ptScaleCenter);
                scaleTransform.CenterX = ptScaleCenter.X;
                scaleTransform.CenterY = ptScaleCenter.Y;

                // Save initial point
                ptPinchStart = args.GetPosition(this);
                viewerDisplayMode = ViewerDisplayMode.Zoomed;
                viewerTouchMode = ViewerTouchMode.Pinching;
            }
            args.Handled = true;
        }

        void OnGestureListenerPinchDelta(object sender, PinchGestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            if (viewerTouchMode == ViewerTouchMode.Pinching)
            {
                // Get raw transform values
                Point ptPosition = args.GetPosition(this);
                double translateX = ptPosition.X - ptPinchStart.X;
                double translateY = ptPosition.Y - ptPinchStart.Y;
                double scale = args.DistanceRatio;

                // Don't allow the page edges to be visible
                ClampPinchAndPanTransforms(scale, translateX, translateY);
            }
            args.Handled = true;
        }

        void OnGestureListenerPinchCompleted(object sender, PinchGestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            if (viewerTouchMode == ViewerTouchMode.Pinching)
            {
                // Transfer Scale and Translate transforms to Matrix transform
                ConsolidateTransforms();
            }
            args.Handled = true;
        }

        void OnGestureListenerGestureCompleted(object sender, GestureEventArgs args)
        {
            if (!gestureAllowed)
                return;

            switch (viewerTouchMode)
            {
                case ViewerTouchMode.Dragging:
                    if (!gestureHadFlick)
                    {
                        if (dragFactor > 0.5)
                            AnimateToNextPage();

                        else if (dragFactor < -0.5)
                            AnimateToPreviousPage();

                        else AnimateToSamePage();
                    }
                    break;

                case ViewerTouchMode.Selecting:
                    // Fire the TextSelected event
                    if (TextSelected != null)
                    {
                        Panel panel = selectedTextBlock1.Parent as Panel;
                        int index1 = panel.Children.IndexOf(selectedTextBlock1);
                        int index2 = panel.Children.IndexOf(selectedTextBlock2);
                        int minIndex = Math.Min(index1, index2);
                        int maxIndex = Math.Max(index1, index2);
                        TextBlock txtblk1 = (panel.Children[minIndex] as TextBlock);
                        TextBlock txtblk2 = (panel.Children[maxIndex] as TextBlock);
                        WordInfo word1 = txtblk1.Tag as WordInfo;
                        WordInfo word2 = txtblk2.Tag as WordInfo;

                        // Get the selected text && its minimum and maximum positions
                        string text = "";
                        Point ulAll = new Point(Double.MaxValue, Double.MaxValue);
                        Point lrAll = new Point(Double.MinValue, Double.MinValue);

                        for (int index = minIndex; index <= maxIndex; index++)
                        {
                            TextBlock txtblk = panel.Children[index] as TextBlock;

                            if (txtblk != null)
                            {
                                text += txtblk.Text;

                                if (index < maxIndex)
                                    text += " ";

                                Point ulText = txtblk.TransformToVisual(this).Transform(new Point());
                                Point lrText = txtblk.TransformToVisual(this).Transform(new Point(txtblk.ActualWidth, txtblk.ActualHeight));

                                ulAll.X = Math.Min(ulAll.X, ulText.X);
                                ulAll.Y = Math.Min(ulAll.Y, ulText.Y);
                                lrAll.X = Math.Max(lrAll.X, lrText.X);
                                lrAll.Y = Math.Max(lrAll.Y, lrText.Y);
                            }
                        }

                        // Set up the event arguments
                        TextSelectedEventArgs eventArgs = new TextSelectedEventArgs
                        {
                            ChapterIndex = this.ChapterIndex,
                            ParagraphIndex = selectedParagraphIndex,
                            CharacterIndex = word1.CharacterIndex,
                            CharacterCount = word2.CharacterIndex - word1.CharacterIndex,
                            Text = text,
                            Bounds = new Rect(ulAll, lrAll)
                        };

                        // Fire the event!
                        TextSelected(this, eventArgs);
                        viewerTouchMode = ViewerTouchMode.Reading;
                    }
                    break;

                case ViewerTouchMode.Pinching:
                case ViewerTouchMode.Panning:
                    viewerTouchMode = ViewerTouchMode.Reading;    // but DisplayMode is still Zoomed!
                    break;
            }

            // Abandon the zoom if the user has restored it to close to actual size
            if (viewerDisplayMode == ViewerDisplayMode.Zoomed)
            {
                if (matrixTransform.Matrix.M11 < 1.05)
                    AnimatePinchAndPanUndo();
            }

            args.Handled = true;
        }
    }
}
