﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using Microsoft.Phone.Controls;

namespace Petzold.Phone.EBookReader
{
    public partial class BookViewer : UserControl
    {
        static readonly TimeSpan pinchUndoAnimationDuration = TimeSpan.FromSeconds(0.5);

        Border[] pageHosts = new Border[3];
        int pageHostBaseIndex = 0;          // current page

        int nextPageChapterIndex;
        int nextPagePageIndex;
        int prevPageChapterIndex;
        int prevPagePageIndex;

        public static readonly DependencyProperty PageProviderProperty =
            DependencyProperty.Register("PageProvider",
                typeof(IPageProvider),
                typeof(BookViewer),
                new PropertyMetadata(OnPageProviderChanged));

        public static readonly DependencyProperty PageTransitionProperty =
            DependencyProperty.Register("PageTransition",
                typeof(PageTransition),
                typeof(BookViewer),
                new PropertyMetadata(OnPageTransitionChanged));

        public static readonly DependencyProperty ChapterIndexProperty =
            DependencyProperty.Register("ChapterIndex",
                typeof(int),
                typeof(BookViewer),
                new PropertyMetadata(0, OnPageChanged));

        public static readonly DependencyProperty PageIndexProperty =
            DependencyProperty.Register("PageIndex",
                typeof(int),
                typeof(BookViewer),
                new PropertyMetadata(0, OnPageChanged));

        // Public events
        public event EventHandler PageChanged;
        public event EventHandler<TextSelectedEventArgs> TextSelected;
        public event EventHandler BookmarkInserted;

        public BookViewer()
        {
            InitializeComponent();

            pageHosts[0] = pageHost0;
            pageHosts[1] = pageHost1;
            pageHosts[2] = pageHost2;
        }

        public IPageProvider PageProvider
        {
            set { SetValue(PageProviderProperty, value); }
            get { return (IPageProvider)GetValue(PageProviderProperty); }
        }

        public PageTransition PageTransition
        {
            set { SetValue(PageTransitionProperty, value); }
            get { return (PageTransition)GetValue(PageTransitionProperty); }
        }

        public int ChapterIndex
        {
            protected set { SetValue(ChapterIndexProperty, value); }
            get { return (int)GetValue(ChapterIndexProperty); }
        }

        public int PageIndex
        {
            protected set { SetValue(PageIndexProperty, value); }
            get { return (int)GetValue(PageIndexProperty); }
        }

        public bool JumpToPage(int chapterIndex, int pageIndex)
        {
            // Save chapter and page indices in public properties
            this.ChapterIndex = chapterIndex;
            this.PageIndex = pageIndex;

            if (this.PageProvider != null)
            {
                bool invalidated = this.PageProvider.SetPageSize(new Size(pageHost0.ActualWidth, pageHost0.ActualHeight));
                invalidated |= this.PageProvider.SetFont(this.FontFamily, this.FontSize);

                // need to return if these have invalidated pagination information
                if (invalidated)
                    return false;

                FrameworkElement element = this.PageProvider.GetPage(chapterIndex, pageIndex);

                if (element != null)
                {
                    // Set pageHosts for this page and the next page
                    pageHosts[pageHostBaseIndex].Child = element;
                    SetUpNextPage();
                    SetUpPreviousPage();

                    // Set the FractionalBaseIndex for the PageTransition
                    this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;
                    return true;
                }
            }
            return false;
        }

        public bool JumpToCharacterOffset(int characterOffset)
        {
            // Ensure that these properties actually change
            this.ChapterIndex = 0;
            this.PageIndex = 0;

            if (this.PageProvider != null)
            {
                this.PageProvider.SetPageSize(new Size(pageHost0.ActualWidth, pageHost0.ActualHeight));
                this.PageProvider.SetFont(this.FontFamily, this.FontSize);
                int chapterIndex;
                int pageIndex;
                FrameworkElement element = this.PageProvider.GetPage(characterOffset, out chapterIndex, out pageIndex);

                if (element != null)
                {
                    this.ChapterIndex = chapterIndex;
                    this.PageIndex = pageIndex;

                    // Set pageHosts for this page and the next page
                    pageHosts[pageHostBaseIndex].Child = element;
                    SetUpNextPage();
                    SetUpPreviousPage();

                    // Set the FractionlBaseIndex for the PageTransition
                    this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;
                    return true;
                }
            }
            return false;
        }

        public void HighlightText(FindInfo findInfo)
        {
            FrameworkElement pageHost = pageHosts[pageHostBaseIndex];
            HighlightText(pageHost, findInfo);
        }

        void HighlightText(FrameworkElement parent, FindInfo findInfo)
        {
            if (parent is TextBlock && (parent as TextBlock).Tag is WordInfo)
            {
                TextBlock txtblk = parent as TextBlock;
                WordInfo wordInfo = txtblk.Tag as WordInfo;

                if (wordInfo.ParagraphIndex == findInfo.ParagraphIndex &&
                    wordInfo.CharacterIndex + wordInfo.CharacterCount > findInfo.StartCharacterIndex &&
                    wordInfo.CharacterIndex <= findInfo.EndCharacterIndex)
                {
                    txtblk.Foreground = this.Resources["PhoneAccentBrush"] as Brush;
                    findInfo.WordLocation = wordInfo.LocationTop;
                }
            }
                    
            int numChildren = VisualTreeHelper.GetChildrenCount(parent);

            for (int i = 0; i < numChildren; i++)
            {
                FrameworkElement child = VisualTreeHelper.GetChild(parent, i) as FrameworkElement;

                if (child != null)
                    HighlightText(child, findInfo);
            }
        }

        public void ClearHighlighting()
        {
            FrameworkElement pageHost = pageHosts[pageHostBaseIndex];
            ClearHighlighting(pageHost);
        }

        void ClearHighlighting(FrameworkElement parent)
        {
            if (parent is TextBlock && (parent as TextBlock).Tag is WordInfo)
            {
                (parent as TextBlock).ClearValue(TextBlock.ForegroundProperty);
            }

            int numChildren = VisualTreeHelper.GetChildrenCount(parent);

            for (int i = 0; i < numChildren; i++)
            {
                FrameworkElement child = VisualTreeHelper.GetChild(parent, i) as FrameworkElement;

                if (child != null)
                    ClearHighlighting(child);
            }
        }

        void SetUpNextPage()
        {
            nextPageChapterIndex = this.ChapterIndex;
            nextPagePageIndex = this.PageIndex + 1;
            FrameworkElement element = this.PageProvider.GetPage(nextPageChapterIndex, nextPagePageIndex);

            if (element == null)
            {
                nextPageChapterIndex += 1;
                nextPagePageIndex = 0;
                element = this.PageProvider.GetPage(nextPageChapterIndex, nextPagePageIndex);
            }

            pageHosts[(pageHostBaseIndex + 1) % 3].Child = element;
        }

        void SetUpPreviousPage()
        {
            prevPageChapterIndex = this.ChapterIndex;
            prevPagePageIndex = this.PageIndex - 1;
            FrameworkElement element = null;

            if (prevPagePageIndex != -1)
            {
                element = this.PageProvider.GetPage(prevPageChapterIndex, prevPagePageIndex);
            }
            else if (prevPageChapterIndex != 0)
            {
                prevPageChapterIndex -= 1;
                element = this.PageProvider.QueryLastPage(prevPageChapterIndex, out prevPagePageIndex);
            }
            pageHosts[(pageHostBaseIndex + 2) % 3].Child = element;
        }

        void ClampPinchAndPanTransforms(double scale, double translateX, double translateY)
        {
            // This is the matrix transform from previous operations
            calcMatrixTransform.Matrix = matrixTransform.Matrix;

            // Calculate scaling factor so it's always 1 or greater
            double totalScale = scale * matrixTransform.Matrix.M11;
            totalScale = Math.Max(1, totalScale);
            scale = totalScale / matrixTransform.Matrix.M11;

            // Set up properties for new scale matrix
            calcScaleTransform.CenterX = scaleTransform.CenterX;
            calcScaleTransform.CenterY = scaleTransform.CenterY;
            calcScaleTransform.ScaleX = scale;
            calcScaleTransform.ScaleY = scale;

            // Set up properties for new translation matrix
            calcTranslateTransform.X = translateX;
            calcTranslateTransform.Y = translateY;

            // Obtain the total matrix from the transform group
            Matrix totalMatrix = calcTransformGroup.Value;

            // Restict translation to original area of transformed element
            double totalTranslationX = totalMatrix.OffsetX;
            double totalTranslationY = totalMatrix.OffsetY;

            double clampedTranslationX = 
                Math.Min(0,
                         Math.Max((1 - totalMatrix.M11) * manipulationGrid.ActualWidth,
                                  totalTranslationX));

            double clampedTranslationY = 
                Math.Min(0,
                         Math.Max((1 - totalMatrix.M22) * manipulationGrid.ActualHeight,
                                  totalTranslationY));

            // Adjust translation factors
            translateX += clampedTranslationX - totalTranslationX;
            translateY += clampedTranslationY - totalTranslationY;

            // Set transforms
            scaleTransform.ScaleX = scale;
            scaleTransform.ScaleY = scale;
            translateTransform.X = translateX;
            translateTransform.Y = translateY;
        }

        void ConsolidateTransforms()
        {
            matrixTransform.Matrix = transformGroup.Value;
            scaleTransform.ScaleX = scaleTransform.ScaleY = 1;
            scaleTransform.CenterX = scaleTransform.CenterY = 0;
            translateTransform.X = translateTransform.Y = 0;
        }

        // Highlight all words in range if select == true; unhighlight all other words on page
        void SelectTextBlockRange(TextBlock txtblk1, TextBlock txtblk2, bool select)
        {
            Panel panel = txtblk1.Parent as Panel;
            int index1 = panel.Children.IndexOf(txtblk1);
            int index2 = panel.Children.IndexOf(txtblk2);
            int minIndex = Math.Min(index1, index2);
            int maxIndex = Math.Max(index1, index2);
            Brush highlightBrush = this.Resources["PhoneAccentBrush"] as Brush;

            for (int index = 0; index < panel.Children.Count; index++)
            {
                TextBlock txtblk = panel.Children[index] as TextBlock;

                if (txtblk != null)
                {
                    if (select && index >= minIndex && index <= maxIndex)
                        txtblk.Foreground = highlightBrush;
                    else
                        txtblk.ClearValue(TextBlock.ForegroundProperty);
                }
            }
        }

        public void AnimateToNextPage()
        {
            DoubleAnimation anima = new DoubleAnimation
            {
                To = pageHostBaseIndex + 1,
                Duration = CalculateFractionalDuration(pageHostBaseIndex + 1)
            };

            Storyboard.SetTarget(anima, this.PageTransition);
            Storyboard.SetTargetProperty(anima, new PropertyPath(PageTransition.FractionalBaseIndexProperty));

            Storyboard storyboard = new Storyboard();
            storyboard.Children.Add(anima);

            storyboard.Completed += (object sender, EventArgs args) =>
            {
                pageHostBaseIndex = (pageHostBaseIndex + 1) % 3;
                this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;

                // Shift the indices from current to previous
                prevPageChapterIndex = this.ChapterIndex;
                prevPagePageIndex = this.PageIndex;

                // Shift the indices from next to current
                this.ChapterIndex = nextPageChapterIndex;
                this.PageIndex = nextPagePageIndex;

                // Set up the new next page
                SetUpNextPage();
                viewerTouchMode = ViewerTouchMode.Reading;
            };

            viewerTouchMode = ViewerTouchMode.Animating;
            storyboard.Begin();
        }

        void AnimateToPreviousPage()
        {
            DoubleAnimation anima = new DoubleAnimation
            {
                To = pageHostBaseIndex - 1,
                Duration = CalculateFractionalDuration(pageHostBaseIndex - 1)
            };

            Storyboard.SetTarget(anima, this.PageTransition);
            Storyboard.SetTargetProperty(anima, new PropertyPath(PageTransition.FractionalBaseIndexProperty));

            Storyboard storyboard = new Storyboard();
            storyboard.Children.Add(anima);

            storyboard.Completed += (object sender, EventArgs args) =>
            {
                pageHostBaseIndex = (pageHostBaseIndex + 2) % 3;
                this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;

                // Transfer the indices from current to next
                nextPageChapterIndex = this.ChapterIndex;
                nextPagePageIndex = this.PageIndex;

                // Transfer the indices from previous to current
                this.ChapterIndex = prevPageChapterIndex;
                this.PageIndex = prevPagePageIndex;

                // Set up the new previous page
                SetUpPreviousPage();
                viewerTouchMode = ViewerTouchMode.Reading;
            };

            viewerTouchMode = ViewerTouchMode.Animating;
            storyboard.Begin();
        }

        void AnimateToSamePage()
        {
            DoubleAnimation anima = new DoubleAnimation
            {
                To = pageHostBaseIndex,
                Duration = CalculateFractionalDuration(pageHostBaseIndex)
            };

            Storyboard.SetTarget(anima, this.PageTransition);
            Storyboard.SetTargetProperty(anima, new PropertyPath(PageTransition.FractionalBaseIndexProperty));

            Storyboard storyboard = new Storyboard();
            storyboard.Children.Add(anima);

            storyboard.Completed += (object sender, EventArgs args) =>
            {
                viewerTouchMode = ViewerTouchMode.Reading;
            };

            viewerTouchMode = ViewerTouchMode.Animating;
            storyboard.Begin();
        }

        public void AnimatePinchAndPanUndo()
        {
            // This should have been done already
            matrixTransform.Matrix = transformGroup.Value;

            // Transfer matrix transform to scale & translate
            scaleTransform.ScaleX = matrixTransform.Matrix.M11;
            scaleTransform.ScaleY = matrixTransform.Matrix.M22;
            scaleTransform.CenterX = 0;
            scaleTransform.CenterY = 0;
            translateTransform.X = matrixTransform.Matrix.OffsetX;
            translateTransform.Y = matrixTransform.Matrix.OffsetY;
            matrixTransform.Matrix = Matrix.Identity;

            Storyboard storyboard = new Storyboard();

            // Scale X Animation
            DoubleAnimation anima = new DoubleAnimation
            {
                To = 1,
                Duration = new Duration(pinchUndoAnimationDuration)
            };

            Storyboard.SetTarget(anima, scaleTransform);
            Storyboard.SetTargetProperty(anima, new PropertyPath(ScaleTransform.ScaleXProperty));
            storyboard.Children.Add(anima);

            // Scale Y Animation
            anima = new DoubleAnimation
            {
                To = 1,
                Duration = new Duration(pinchUndoAnimationDuration)
            };

            Storyboard.SetTarget(anima, scaleTransform);
            Storyboard.SetTargetProperty(anima, new PropertyPath(ScaleTransform.ScaleYProperty));
            storyboard.Children.Add(anima);

            // Translate X Animation
            anima = new DoubleAnimation
            {
                To = 0,
                Duration = new Duration(pinchUndoAnimationDuration)
            };

            Storyboard.SetTarget(anima, translateTransform);
            Storyboard.SetTargetProperty(anima, new PropertyPath(TranslateTransform.XProperty));
            storyboard.Children.Add(anima);

            // Translate Y Animation
            anima = new DoubleAnimation
            {
                To = 0,
                Duration = new Duration(pinchUndoAnimationDuration)
            };

            Storyboard.SetTarget(anima, translateTransform);
            Storyboard.SetTargetProperty(anima, new PropertyPath(TranslateTransform.YProperty));
            storyboard.Children.Add(anima);

            // Set Completed handler
            storyboard.Completed += delegate
            {
                viewerDisplayMode = ViewerDisplayMode.Normal;
                viewerTouchMode = ViewerTouchMode.Reading;
            };

            // Start the animation
            viewerTouchMode = ViewerTouchMode.Animating;
            storyboard.Begin();
        }

        Duration CalculateFractionalDuration(double fractionalBaseIndexDestination)
        {
            double fraction = Math.Abs(fractionalBaseIndexDestination -
                                       this.PageTransition.FractionalBaseIndex);
            double msec = fraction * this.PageTransition.AnimationDuration;
            TimeSpan timeSpan = TimeSpan.FromMilliseconds(msec);
            return new Duration(timeSpan);
        }

        static void OnPageProviderChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
        }

        static void OnPageTransitionChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as BookViewer).OnPageTransitionChanged((PageTransition)args.OldValue, (PageTransition)args.NewValue);
        }

        void OnPageTransitionChanged(PageTransition oldPageTransition, PageTransition newPageTransition)
        {
            double fractionalBaseIndex = -1;

            if (oldPageTransition != null)
            {
                fractionalBaseIndex = oldPageTransition.FractionalBaseIndex;
                oldPageTransition.Detach();
            }
            if (newPageTransition != null)
            {
                newPageTransition.Attach(LayoutRoot, pageContainer0, pageContainer1, pageContainer2);
                newPageTransition.FractionalBaseIndex = fractionalBaseIndex;
            }
        }

        static void OnPageChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            BookViewer bookViewer = obj as BookViewer;

            if (bookViewer.PageChanged != null)
                bookViewer.PageChanged(bookViewer, EventArgs.Empty);
        }
    }
}
