﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Media;

namespace Petzold.Phone.EBookReader
{
    [TemplatePart(Name = "HorizontalTemplate", Type = typeof(FrameworkElement))]
    [TemplatePart(Name = "HorizontalTrack", Type = typeof(Border))]
    [TemplatePart(Name = "HorizontalThumb", Type = typeof(ManipulationThumb))]
    [TemplatePart(Name = "HorizontalThumbTranslate", Type = typeof(TranslateTransform))]
    [TemplatePart(Name = "VerticalTemplate", Type = typeof(FrameworkElement))]
    [TemplatePart(Name = "VerticalTrack", Type = typeof(Border))]
    [TemplatePart(Name = "VerticalThumb", Type = typeof(ManipulationThumb))]
    [TemplatePart(Name = "VerticalThumbTranslate", Type = typeof(TranslateTransform))]
    public class IntegralSlider : Control
    {
        FrameworkElement horizontalTemplate, verticalTemplate;
        Border horizontalTrack, verticalTrack, activeTrack;
        ManipulationThumb horizontalThumb, verticalThumb, activeThumb;
        TranslateTransform horizontalThumbTranslate, verticalThumbTranslate, activeThumbTranslate;
        bool isDragging;
        double trackSize;
        double rawOffset;

        public static readonly DependencyProperty OrientationProperty =
            DependencyProperty.Register("Orientation",
                typeof(Orientation),
                typeof(IntegralSlider),
                new PropertyMetadata(Orientation.Horizontal, OnOrientationChanged));

        public static readonly DependencyProperty MinimumProperty =
            DependencyProperty.Register("Minimum",
                typeof(int),
                typeof(IntegralSlider),
                new PropertyMetadata(0, OnMinMaxChanged));

        public static readonly DependencyProperty MaximumProperty =
            DependencyProperty.Register("Maximum",
                typeof(int),
                typeof(IntegralSlider),
                new PropertyMetadata(10, OnMinMaxChanged));

        public static readonly DependencyProperty ValueProperty =
            DependencyProperty.Register("Value",
                typeof(int),
                typeof(IntegralSlider),
                new PropertyMetadata(0, OnValueChanged));

        public event RoutedPropertyChangedEventHandler<int> ValueChanged;

        public IntegralSlider()
        {
            this.DefaultStyleKey = typeof(IntegralSlider);
            Loaded += (sender, args) =>
            {
                MoveThumbToPosition();
            };
        }

        public double SmallThumbDimension { get { return 48; } }
        public double LargeThumbDimension { get { return 96; } }

        public Orientation Orientation
        {
            set { SetValue(OrientationProperty, value); }
            get { return (Orientation)GetValue(OrientationProperty); }
        }

        public int Minimum
        {
            set { SetValue(MinimumProperty, value); }
            get { return (int)GetValue(MinimumProperty); }
        }

        public int Maximum
        {
            set { SetValue(MaximumProperty, value); }
            get { return (int)GetValue(MaximumProperty); }
        }

        public int Value
        {
            set { SetValue(ValueProperty, value); }
            get { return (int)GetValue(ValueProperty); }
        }

        static void OnOrientationChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as IntegralSlider).OnOrientationChanged(args);
        }

        void OnOrientationChanged(DependencyPropertyChangedEventArgs args)
        {
            SwapOrientation();
        }

        static void OnMinMaxChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as IntegralSlider).OnMinMaxChanged(args);
        }

        void OnMinMaxChanged(DependencyPropertyChangedEventArgs args)
        {
            if ((Value < Minimum && Minimum <= Maximum) ||
                (Value > Minimum && Minimum >= Maximum))
            {
                Value = Minimum;
            }

            else if ((Value < Maximum && Maximum <= Minimum) ||
                     (Value > Maximum && Maximum >= Minimum))
            {
                Value = Maximum;
            }

            if (!isDragging)
                MoveThumbToPosition();
        }

        static void OnValueChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as IntegralSlider).OnValueChanged((int)args.OldValue, (int)args.NewValue);
        }

        protected virtual void OnValueChanged(int oldValue, int newValue)
        {
            if ((Minimum <= Maximum && Value >= Minimum && Value <= Maximum) ||
                (Maximum <= Minimum && Value >= Maximum && Value <= Minimum))
            {
                if (ValueChanged != null)
                    ValueChanged(this, new RoutedPropertyChangedEventArgs<int>(oldValue, newValue));

                if (!isDragging)
                {
                    MoveThumbToPosition();
                }
            }
            else
            {
                OnMinMaxChanged(new DependencyPropertyChangedEventArgs());
            }
        }

        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();

            horizontalTemplate = this.GetTemplateChild("HorizontalTemplate") as FrameworkElement;
            horizontalTrack = this.GetTemplateChild("HorizontalTrack") as Border;
            horizontalThumb = this.GetTemplateChild("HorizontalThumb") as ManipulationThumb;
            horizontalThumbTranslate = this.GetTemplateChild("HorizontalThumbTranslate") as TranslateTransform;

            verticalTemplate = this.GetTemplateChild("VerticalTemplate") as FrameworkElement;
            verticalTrack = this.GetTemplateChild("VerticalTrack") as Border;
            verticalThumb = this.GetTemplateChild("VerticalThumb") as ManipulationThumb;
            verticalThumbTranslate = this.GetTemplateChild("VerticalThumbTranslate") as TranslateTransform;

            SwapOrientation();
            MoveThumbToPosition();
        }

        void SwapOrientation()
        {
            if (activeThumb != null && activeThumbTranslate != null)
            {
                activeThumb.DragStarted -= OnThumbDragStarted;
                activeThumb.DragDelta -= OnThumbDragDelta;
                activeThumb.DragCompleted -= OnThumbDragCompleted;
            }

            if (Orientation == Orientation.Horizontal)
            {
                if (horizontalTemplate != null)
                    horizontalTemplate.Visibility = Visibility.Visible;

                if (verticalTemplate != null)
                    verticalTemplate.Visibility = Visibility.Collapsed;

                activeTrack = horizontalTrack;
                activeThumb = horizontalThumb;
                activeThumbTranslate = horizontalThumbTranslate;
            }
            else
            {
                if (horizontalTemplate != null)
                    horizontalTemplate.Visibility = Visibility.Collapsed;

                if (verticalTemplate != null)
                    verticalTemplate.Visibility = Visibility.Visible;

                activeTrack = verticalTrack;
                activeThumb = verticalThumb;
                activeThumbTranslate = verticalThumbTranslate;
            }

            if (activeThumb != null && activeThumbTranslate != null)
            {
                activeThumb.DragStarted += OnThumbDragStarted;
                activeThumb.DragDelta += OnThumbDragDelta;
                activeThumb.DragCompleted += OnThumbDragCompleted;
            }
            base.OnApplyTemplate();
        }

        void MoveThumbToPosition()
        {
            if (activeThumb == null || activeThumbTranslate == null)
                return;

            double trackSize = GetTrackSize();
            double offset = ((double)(Value - Minimum) / Math.Abs(Maximum - Minimum)) * trackSize;

            if (Orientation == Orientation.Horizontal)
                activeThumbTranslate.X = offset;
            else
                activeThumbTranslate.Y = offset;
        }

        double GetTrackSize()
        {
            FrameworkElement container = (activeTrack as FrameworkElement) ??
                                         (this as FrameworkElement);

            if (Orientation == Orientation.Horizontal)
                return container.ActualWidth - activeThumb.Width;

            return container.ActualHeight - activeThumb.Height;
        }

        void OnThumbDragStarted(object sender, DragStartedEventArgs args)
        {
            isDragging = true;

            FrameworkElement container = (activeTrack as FrameworkElement) ??
                                         (this as FrameworkElement);

            if (Orientation == Orientation.Horizontal)
            {
                trackSize = container.ActualWidth - activeThumb.Width;
                rawOffset = horizontalThumbTranslate.X;
            }
            else
            {
                trackSize = container.ActualHeight - activeThumb.Height;
                rawOffset = verticalThumbTranslate.Y;
            }
        }

        void OnThumbDragDelta(object sender, DragDeltaEventArgs args)
        {
            if (!isDragging)
                return;

            if (Orientation == Orientation.Horizontal)
                rawOffset += args.HorizontalChange;
            else
                rawOffset += args.VerticalChange;

            rawOffset = Math.Max(0, Math.Min(trackSize, rawOffset));
            double useOffset = rawOffset - (trackSize / 35) * Math.Sin(2 * Math.PI * (rawOffset / trackSize) * Math.Abs(Maximum - Minimum));

            if (Orientation == Orientation.Horizontal)
                activeThumbTranslate.X = useOffset;
            else
                activeThumbTranslate.Y = useOffset;

            Value = (int)Math.Round(Minimum + useOffset / trackSize * (Maximum - Minimum));
        }

        void OnThumbDragCompleted(object sender, DragCompletedEventArgs args)
        {
            isDragging = false;
        }
    }
}
