﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace Petzold.Phone.EBookReader
{
    public class PageProvider : IPageProvider
    {
        const int FirstLineIndent = 3;              // in units of space width
        const double ParagraphSpacing = 1.0 / 3;    // in units of character height

        FontMetrics fontMetrics = new FontMetrics(new Font());
        double fontSize;
        Size pageSize;

        public PageProvider(BookInfo book)
        {
            this.Book = book;
        }

        // Only public property
        public BookInfo Book { protected set; get; }

        public bool SetPageSize(Size size)
        {
            bool invalidate = false;
            this.pageSize = size;

            if (this.Book.PaginationPageSize != pageSize)
            {
                InvalidatePaginationInformation();
                this.Book.PaginationPageSize = pageSize;
                invalidate = true;
            }
            return invalidate;
        }

        public bool SetFont(FontFamily fontFamily, double fontSize)
        {
            if (fontFamily != fontMetrics.Font.FontFamily)
                this.fontMetrics = new FontMetrics(new Font(fontFamily.Source));

            bool invalidate = false;
            this.fontSize = fontSize;

            if (this.Book.PaginationFontFamily != fontFamily.Source ||
                this.Book.PaginationFontSize != fontSize)
            {
                InvalidatePaginationInformation();
                this.Book.PaginationFontFamily = fontFamily.Source;
                this.Book.PaginationFontSize = fontSize;
                invalidate = true;
            }
            return invalidate;
        }

        void InvalidatePaginationInformation()
        {
            foreach (ChapterInfo chapter in this.Book.Chapters)
            {
                int pageCount = chapter.Pages.Count;

                if (pageCount > 1)
                    chapter.Pages.RemoveRange(1, pageCount - 1);

                chapter.Pages[0].IsLastPageInChapter = false;
                chapter.Pages[0].IsPaginated = false;
                chapter.Pages[0].Words.Clear();
            }

            this.Book.IsChapterAndPageInvalid = true;
        }

        public FrameworkElement GetPage(int characterCount, out int chapterIndex, out int pageIndex)
        {
            chapterIndex = 0;
            pageIndex = 0;

            for (int ch = 0; ch < this.Book.Chapters.Count; ch++)
            {
                if (this.Book.Chapters[ch].Pages[0].AccumulatedCharacterCount <= characterCount)
                    chapterIndex = ch;
                else
                    break;
            }

            for (int pg = 0; ; pg++)
            {
                bool success = Paginate(chapterIndex, pg);

                if (success && this.Book.Chapters[chapterIndex].Pages[pg].AccumulatedCharacterCount <= characterCount)
                {
                    pageIndex = pg;
                }
                else
                    break;
            }
            return GetPage(chapterIndex, pageIndex);
        }

        public FrameworkElement GetPage(int chapterIndex, int pageIndex)
        {
            if (pageSize.IsEmpty)
                return null;

            if (pageIndex < 0)
                return null;

            if (chapterIndex < 0 || chapterIndex > this.Book.Chapters.Count - 1)
                return null;

            ChapterInfo chapter = this.Book.Chapters[chapterIndex];

            // Check if the last page of the paragraph has already been established, 
            //  and the requested page is greater than that last page
            if (chapter.Pages[chapter.Pages.Count - 1].IsLastPageInChapter && pageIndex > chapter.Pages.Count - 1)
                return null;

            // If the last page of a chapter is requested, and it's already available, get it
            if (pageIndex < chapter.Pages.Count && chapter.Pages[pageIndex].IsLastPageInChapter)
            {
                return BuildPageElement(chapter, chapter.Pages[pageIndex]);
            }

            // Otherwise, if the page has already been paginated, get it
            if (pageIndex + 1 < chapter.Pages.Count && chapter.Pages[pageIndex].IsPaginated)
            {
                return BuildPageElement(chapter, chapter.Pages[pageIndex]);
            }

            if (pageIndex != chapter.Pages.Count - 1)
                throw new Exception(String.Format("GetPage can only paginate one page per call: pageIndex = {0} while chapter.Pages.Count = {1}", pageIndex, chapter.Pages.Count));

            PageInfo pageInfo = chapter.Pages[pageIndex];

            PageInfo nextPageInfo = Paginate(chapter, pageInfo);

            if (nextPageInfo != null)   // ie, if !pageInfo.IsLastPageInChapter
                chapter.Pages.Add(nextPageInfo);

            return BuildPageElement(chapter, pageInfo);
        }

        public FrameworkElement QueryLastPage(int chapterIndex, out int pageIndex) 
        {
            return GetLastPage(chapterIndex, out pageIndex, false);
        }

        public FrameworkElement GetLastPage(int chapterIndex, out int pageIndex)
        {
            FrameworkElement element = GetLastPage(chapterIndex, out pageIndex, true);
            return element;
        }

        FrameworkElement GetLastPage(int chapterIndex, out int pageIndex, bool paginateIfNecessary)
        {
            if (pageSize.IsEmpty)
            {
                pageIndex = -1;
                return null;
            }

            if (chapterIndex < 0 || chapterIndex > this.Book.Chapters.Count - 1)
            {
                pageIndex = -1;
                return null;
            }

            ChapterInfo chapter = this.Book.Chapters[chapterIndex];
            int lastIndex = chapter.Pages.Count - 1;

            if (chapter.Pages[lastIndex].IsLastPageInChapter && 
                chapter.Pages[lastIndex].IsPaginated)               // IsPaginated should be implied
            {
                pageIndex = lastIndex;
                return BuildPageElement(chapter, chapter.Pages[lastIndex]);
            }

            if (!paginateIfNecessary)
            {
                pageIndex = -1;
                return null;
            }

            while (!chapter.Pages[chapter.Pages.Count - 1].IsLastPageInChapter)
            {
                Paginate(chapterIndex, chapter.Pages.Count - 1);
            }

            pageIndex = chapter.Pages.Count - 1;
            return BuildPageElement(chapter, chapter.Pages[chapter.Pages.Count - 1]);
        }

        bool Paginate(int chapterIndex, int pageIndex)
        {
            if (chapterIndex < 0 || chapterIndex > this.Book.Chapters.Count - 1)
                return false;

            ChapterInfo chapter = this.Book.Chapters[chapterIndex];

            if (pageIndex < chapter.Pages.Count && chapter.Pages[pageIndex].IsPaginated)
                return true;

            if (chapter.Pages[chapter.Pages.Count - 1].IsLastPageInChapter)
                return false;

            PageInfo pageInfo = chapter.Pages[pageIndex];
            PageInfo nextPageInfo = Paginate(chapter, pageInfo);

            if (nextPageInfo != null)
                chapter.Pages.Add(nextPageInfo);

            return true;
        }

        // TODO: Could pass chapter and pageIndex here, to obtain pageInfo within, and then can set nextPageInfo
        PageInfo Paginate(ChapterInfo chapter, PageInfo pageInfo)
        {
            // Get dimensions of space character
            Size spaceSize = fontMetrics.MeasureText(" ");
            int lineHeight = (int)Math.Ceiling(fontSize * spaceSize.Height);
            int spaceWidth = (int)Math.Ceiling(fontSize * spaceSize.Width);

            // Initialize a couple variables
            int accumHeight = 0;
            int characterCount = pageInfo.AccumulatedCharacterCount;
            PageInfo nextPageInfo = null;
            int paragraphIndex = pageInfo.ParagraphIndex;
            int characterIndex = pageInfo.CharacterIndex;

            // Loop for lines on page
            while (true)
            {
                ParagraphInfo paragraph = this.Book.Paragraphs[paragraphIndex];

                // For blank paragraphs, leave a gap...
                if (paragraph.Text.Length == 0)
                {
                    // .. but not when it's at the top of the page
                    if (accumHeight > 0)
                        accumHeight += (int)Math.Ceiling(ParagraphSpacing * lineHeight);
                }
                // Normal case for non-blank paragraphs
                else
                {
                    // This is the amount to indent the entire paragraph
                    int indent = spaceWidth * paragraph.Indent;

                    // Apply first line indent if paragraph indent is zero, and it's
                    //  not the first paragraph in the chapter
                    if (indent == 0 &&
                        characterIndex == 0 &&
                        paragraph != this.Book.Paragraphs[chapter.FirstParagraph])
                    {
                        indent = spaceWidth * FirstLineIndent;
                    }

                    // Start off accumulated line width with the indent
                    int accumWidth = indent;

                    // Skip blanks that might be at the beginning of a line.
                    //  (This can happen if sentences are separated by two
                    //      blanks, and a sentence ends in the preceding line.)
                    while (paragraph.Text[characterIndex] == ' ')
                        characterIndex += 1;

                    // Loop for words within line -- always the same paragraph
                    while (true)
                    {
                        // Find next space in paragraph
                        int spaceIndex = paragraph.Text.IndexOf(' ', characterIndex);
                        int wordCharacters = (spaceIndex == -1 ? paragraph.Text.Length : spaceIndex) - characterIndex;
                        int wordWidth = (int)Math.Ceiling(fontSize * fontMetrics.MeasureText(paragraph.Text, characterIndex, wordCharacters).Width);

                        // Check if the accumulated width exceeds the page width, 
                        //      but make sure to have at least one word.
                        if (accumWidth > indent && accumWidth + wordWidth >= pageSize.Width)
                            break;

                        WordInfo word = new WordInfo
                        {
                            LocationLeft = accumWidth, 
                            LocationTop = accumHeight,
                            ParagraphIndex = paragraphIndex,
                            CharacterIndex = characterIndex,
                            CharacterCount = wordCharacters,
                        };
                        pageInfo.Words.Add(word);

                        accumWidth += wordWidth;
                        characterCount += wordCharacters + (spaceIndex == -1 ? 0 : 1);
                        characterIndex += wordCharacters + (spaceIndex == -1 ? 0 : 1);

                        // End of line because it's the end of the paragraph
                        if (spaceIndex == -1)
                        {
                            characterIndex = 0;
                            break;
                        }

                        // End of line because no more words will fit
                        if (accumWidth + spaceWidth >= pageSize.Width)
                            break;

                        accumWidth += spaceWidth;

                        // Otherwise, point to next word
                        characterIndex = spaceIndex + 1;

                        // ready for next word
                    }
                    // Bump up the accumulated height of the page
                    accumHeight += lineHeight;
                }

                // Start a new line, but check first if it's a new paragraph
                if (characterIndex == 0)
                {
                    // If we've been working with the last paragraph in the chapter,
                    //      we're also done with the page
                    if (paragraphIndex + 1 == chapter.FirstParagraph + chapter.ParagraphCount)
                    {
                        pageInfo.IsLastPageInChapter = true;
                        break;
                    }
                    // Otherwise, kick up the paragraphIndex
                    paragraphIndex++;
                }

                // Check if the next paragraph is too much for this page
                if ((this.Book.Paragraphs[paragraphIndex].Text.Length == 0 &&
                            accumHeight + lineHeight / 2 > pageSize.Height) ||
                    accumHeight + lineHeight > pageSize.Height)
                {
                    break;
                }
            }

            pageInfo.IsPaginated = true;

            // Create PageInfo for next page
            if (!pageInfo.IsLastPageInChapter)
            {
                nextPageInfo = new PageInfo
                {
                    ParagraphIndex = paragraphIndex,
                    CharacterIndex = characterIndex,
                    AccumulatedCharacterCount = characterCount
                };
            }

            return nextPageInfo;
        }

        FrameworkElement BuildPageElement(ChapterInfo chapter, PageInfo pageInfo)
        {
            if (pageInfo.Words.Count == 0)
            {
                Paginate(chapter, pageInfo);
            }

            Canvas canvas = new Canvas();

            foreach (WordInfo word in pageInfo.Words)
            {
                TextBlock txtblk = new TextBlock
                {
                    FontFamily = fontMetrics.Font.FontFamily,
                    FontSize = this.fontSize,
                    Text = this.Book.Paragraphs[word.ParagraphIndex].Text.
                                Substring(word.CharacterIndex, 
                                          word.CharacterCount),
                    Tag = word
                };

                Canvas.SetLeft(txtblk, word.LocationLeft);
                Canvas.SetTop(txtblk, word.LocationTop);
                canvas.Children.Add(txtblk);
            }

            // Avoid heavy build-up of WordInfo collections
            pageInfo.Words.Clear();

            return canvas;
        }

        public FindInfo FindText(int paragraphIndex, int characterIndex,
                         string text, bool matchCase, bool wholeWord, bool goUp)
        {
            if (paragraphIndex < 0 || paragraphIndex >= this.Book.Paragraphs.Count)
                return null;

            StringComparison comparisonType = matchCase ? StringComparison.Ordinal : StringComparison.OrdinalIgnoreCase;

            if (!goUp)
            {
                return FindTextDown(paragraphIndex, characterIndex, text, wholeWord, comparisonType);
            }
            else
            {
                return FindTextUp(paragraphIndex, characterIndex, text, wholeWord, comparisonType);
            }
        }

        FindInfo FindTextDown(int paragraphIndex, int characterIndex, string text, bool wholeWord, StringComparison comparisonType)
        {
            while (paragraphIndex < this.Book.Paragraphs.Count)
            {
                string paragraphText = this.Book.Paragraphs[paragraphIndex].Text;

                while (characterIndex + text.Length <= paragraphText.Length)
                {
                    int findIndex = paragraphText.IndexOf(text, characterIndex, comparisonType);

                    if (findIndex == -1)
                    {
                        break;
                    }
                    else if (!wholeWord || TestWholeWord(paragraphText, findIndex, text.Length))
                    {
                        return CreateFindInfo(paragraphIndex, findIndex, text.Length);
                    }
                    else
                    {
                        characterIndex = findIndex + 1;
                    }
                }

                // Go to the next paragraph
                paragraphIndex += 1;
                characterIndex = 0;
            }
            return null;
        }

        FindInfo FindTextUp(int paragraphIndex, int characterIndex, string text, bool wholeWord, StringComparison comparisonType)
        {
            characterIndex = Math.Min(this.Book.Paragraphs[paragraphIndex].Text.Length - 1, characterIndex + text.Length - 2);

            while (paragraphIndex >= 0)
            {
                string paragraphText = this.Book.Paragraphs[paragraphIndex].Text;

                while (characterIndex > text.Length - 2)
                {
                    int findIndex = paragraphText.LastIndexOf(text, characterIndex, comparisonType);

                    if (findIndex == -1)
                    {
                        break;
                    }
                    else if (!wholeWord || TestWholeWord(paragraphText, findIndex, text.Length))
                    {
                        return CreateFindInfo(paragraphIndex, findIndex, text.Length);
                    }
                    else
                    {
                        characterIndex = findIndex - 1;
                    }
                }

                // Go to the previous paragraph
                paragraphIndex -= 1;

                if (paragraphIndex >= 0)
                    characterIndex = this.Book.Paragraphs[paragraphIndex].Text.Length - 1;
            }
            return null;
        }

        bool TestWholeWord(string paragraphText, int findIndex, int findTextLength)
        {
            // If wholeWord is true, then the matched text must be either 
            //      at the beginning of the paragraph or be prefaced
            //      by a space or punctuation symbol AND
            //      it must either be at the end of the paragraph or be
            //      followed by a space or punctuation.
            int prevCharIndex = findIndex - 1;
            int nextCharIndex = findIndex + findTextLength;
            bool isMatch = (prevCharIndex == -1 ||
                                paragraphText[prevCharIndex] == ' ' ||
                                    Char.IsPunctuation(paragraphText[prevCharIndex])) &&
                           (nextCharIndex == paragraphText.Length ||
                                paragraphText[nextCharIndex] == ' ' ||
                                    Char.IsPunctuation(paragraphText[nextCharIndex]));
            return isMatch;
        }

        FindInfo CreateFindInfo(int paragraphIndex, int characterIndex, int findTextLength)
        {
            int characterOffset = this.Book.Paragraphs[paragraphIndex].AccumulatedCharacterCount + characterIndex;

            FindInfo findInfo = new FindInfo
            {
                CharacterOffset = characterOffset,
                ParagraphIndex = paragraphIndex,
                StartCharacterIndex = characterIndex,
                EndCharacterIndex = characterIndex + findTextLength     // should this be used directly????
            };

            return findInfo;
        }
    }
}

